@@ -3,15 +3,16 @@ from binascii import unhexlify
 from datetime import datetime
 from itertools import islice
 import errno
-import logging
-logger = logging.getLogger(__name__)
-
 import os
 import shutil
 import struct
 from zlib import crc32
 
 import msgpack
+
+from .logger import create_logger
+logger = create_logger()
+
 from .helpers import Error, ErrorWithTraceback, IntegrityError, Location, ProgressIndicatorPercent, bin_to_hex
 from .hashindex import NSIndex
 from .locking import Lock, LockError, LockErrorT
@@ -438,6 +439,9 @@ class Repository:
             transaction_id = self.get_index_transaction_id()
         if transaction_id is None:
             transaction_id = self.io.get_latest_segment()
+        if transaction_id is None:
+            report_error('This repository contains no valid data.')
+            return False
         if repair:
             self.io.cleanup(transaction_id)
         segments_transaction_id = self.io.get_segments_transaction_id()
@@ -666,7 +670,8 @@ class LoggedIO:
                 if not os.path.exists(dirname):
                     os.mkdir(dirname)
                     sync_dir(os.path.join(self.path, 'data'))
-            self._write_fd = open(self.segment_filename(self.segment), 'ab')
+            # play safe: fail if file exists (do not overwrite existing contents, do not append)
+            self._write_fd = open(self.segment_filename(self.segment), 'xb')
             self._write_fd.write(MAGIC)
             self.offset = MAGIC_LEN
         return self._write_fd
@@ -816,18 +821,35 @@ class LoggedIO:
         self.close_segment()  # after-commit fsync()
 
     def close_segment(self):
-        if self._write_fd:
-            self.segment += 1
-            self.offset = 0
-            self._write_fd.flush()
-            os.fsync(self._write_fd.fileno())
-            if hasattr(os, 'posix_fadvise'):  # only on UNIX
-                # tell the OS that it does not need to cache what we just wrote,
-                # avoids spoiling the cache for the OS and other processes.
-                os.posix_fadvise(self._write_fd.fileno(), 0, 0, os.POSIX_FADV_DONTNEED)
-            self._write_fd.close()
-            sync_dir(os.path.dirname(self._write_fd.name))
-            self._write_fd = None
+        # set self._write_fd to None early to guard against reentry from error handling code pathes:
+        fd, self._write_fd = self._write_fd, None
+        if fd is not None:
+            dirname = None
+            try:
+                self.segment += 1
+                self.offset = 0
+                dirname = os.path.dirname(fd.name)
+                fd.flush()
+                fileno = fd.fileno()
+                os.fsync(fileno)
+                if hasattr(os, 'posix_fadvise'):  # only on UNIX
+                    try:
+                        # tell the OS that it does not need to cache what we just wrote,
+                        # avoids spoiling the cache for the OS and other processes.
+                        os.posix_fadvise(fileno, 0, 0, os.POSIX_FADV_DONTNEED)
+                    except OSError:
+                        # usually, posix_fadvise can't fail for us, but there seem to
+                        # be failures when running borg under docker on ARM, likely due
+                        # to a bug outside of borg.
+                        # also, there is a python wrapper bug, always giving errno = 0.
+                        # https://github.com/borgbackup/borg/issues/2095
+                        # as this call is not critical for correct function (just to
+                        # optimize cache usage), we ignore these errors.
+                        pass
+            finally:
+                fd.close()
+                if dirname:
+                    sync_dir(dirname)
 
 
 MAX_DATA_SIZE = MAX_OBJECT_SIZE - LoggedIO.put_header_fmt.size
